import os
import re
import pprint
import webbrowser
import platform
import subprocess

from qtpy import QtWidgets, QtGui, QtCore

from libreflow_launcher.ui import icons
import libreflow_launcher.resources as resources


class LabelIcon(QtWidgets.QLabel):

    def __init__(self, icon=None, size=None, color=None):
        QtWidgets.QLabel.__init__(self, '')
        if icon:
            self.setIcon(icon, size, color)
    
    def setIcon(self, icon, size=None, color=None):
        icon = QtGui.QIcon(resources.get_icon(icon, color=color))
        if size:
            size_value = QtCore.QSize(size, size)
        else:
            size_value = QtCore.QSize(16, 16)
        pixmap = icon.pixmap(size_value)
        self.setPixmap(pixmap)
        self.setAlignment(QtCore.Qt.AlignVCenter)


class LineEdit(QtWidgets.QLineEdit):

    def __init__(self, value=None, options=[], config_key=None, parent_widget=None):
        QtWidgets.QLineEdit.__init__(self)
        if value:
            self.setText(value)
        
        self.base_value = self.text()
        self.config_key = config_key
        self.options = options
        self.parent_widget = parent_widget

        if "path" in self.options:
            self.reveal_explorer()
        
        if "host" not in self.options:
            self.textChanged.connect(self._on_text_changed)

        self.editingFinished.connect(self._on_text_finish_edit)
        self.returnPressed.connect(self._on_enter_pressed)

    def reset(self):
        self.base_value = self.text()
        self.setProperty('edited', False)
        self.style().polish(self)

    def value_check(self):
        if self.text() == '':
            if "mandatory" in self.options:
                self.setProperty('warning', True)
                self.setToolTip('You cannot leave this parameter empty.')
                return False
        
        if "host" in self.options:
            host_regex = r'^(?:\b(?!\bhttp[s]{0,1})([a-zA-Z\.\/]+)|(\d+\.\d+.\d+.\d+))(?::(\d{1,5}))?'
            if re.match(host_regex, self.text()) is None:
                self.setProperty('warning', True)
                self.setToolTip('Host format is not valid.')
                return False

        return True

    def reveal_explorer(self):
        self.open_path = QtWidgets.QToolButton(self)
        self.open_path.setIcon(QtGui.QIcon(resources.get_icon(('icons.gui', 'open'))))
        self.open_path.setIconSize(QtCore.QSize(15,15))
        self.open_path.setStyleSheet('border: 0px; padding: 0px;')
        self.open_path.setCursor(QtCore.Qt.ArrowCursor)
        self.open_path.clicked.connect(self._on_explorer_button_clicked)

        layout = QtWidgets.QHBoxLayout(self)
        layout.addWidget(self.open_path, 0, QtCore.Qt.AlignRight)

        layout.setSpacing(0)
        layout.setContentsMargins(5,5,5,5)

    def _on_text_changed(self):
        checked = self.value_check()
        if checked:
            if self.text() != self.base_value:
                self.setProperty('warning', False)
                self.setProperty('edited', True)
            else:
                self.setProperty('warning', False)
                self.setProperty('edited', False)
            self.setToolTip('')
        
        self.style().polish(self)
        if self.config_key:
            self.parent_widget._page.toggle_footer()

    def _on_text_finish_edit(self):
        self._on_text_changed()

    def _on_enter_pressed(self):
        self._on_text_finish_edit()
        self.parent().focusNextChild()
        focused_widget = QtWidgets.QApplication.focusWidget()
        if focused_widget.property('default'):
            focused_widget.clicked.emit()

    def _on_explorer_button_clicked(self):
        path = self.parent()._ctrl.resolve_value(self.text())
        if '{}' in path:
            path = path.split('{}')[0]
            self.parent()._mw.notification_header.pop("Open the most accessible folder", 'info', True)

        if os.path.isdir(path):
            if platform.system() == 'Windows':
                return os.startfile(path)
            elif platform.system() == 'Linux':
                return subprocess.Popen(f'xdg-open "{path}"')
            else:
                return subprocess.Popen(f'open "{path}"')
        
        return self.parent()._mw.notification_header.pop("Path is not valid", 'error', True)


class ComboBox(QtWidgets.QComboBox):

    def __init__(self, items=None, config_key=None, parent=None):
        QtWidgets.QComboBox.__init__(self)
        if items:
            self.addItems(items)
        
        self.config_key = config_key
        self.parent = parent

    def enable_check(self):
        self.base_value = self.currentIndex()
        self.currentIndexChanged.connect(self._on_index_changed)
    
    def reset(self):
        self.base_value = self.currentIndex()
        self.setProperty('edited', False)
        self.style().polish(self)

    def _on_index_changed(self, i):
        if i != self.base_value:
            self.setProperty('edited', True)
        else:
            self.setProperty('edited', False)
        
        self.style().polish(self)
        self.parent._page.toggle_footer()


class MainWindow(QtWidgets.QWidget):

    def __init__(self, ctrl):
        QtWidgets.QWidget.__init__(self)
        # self._model = model
        self._ctrl = ctrl
        self.overlay_enabled = False

        self.setWindowIcon(QtGui.QIcon(resources.get_icon(('icons.gui', 'feespeciales_icon'))))
        self.resize(800, 600)
        
        vlo = QtWidgets.QVBoxLayout(self)
        vlo.setContentsMargins(0,0,0,0)

        background = QtWidgets.QFrame()
        background.setObjectName('Background')
        background.setStyleSheet(
            '''
            #Background {
                background-color: white;
            }
            '''
        )
        vlo.addWidget(background)

        self.grid_lo = QtWidgets.QGridLayout(background)
        self.grid_lo.setContentsMargins(0,0,0,0)

        self.menu_bar = MenuBar(self)
        self.wizard = WizardPage(self)
        self.login = LoginPage(self)
        self.projects = ProjectsList(self)
        self.settings_bar = SettingsBar(self)
        self.settings_bar.hide()
        self.settings_page = SettingsPage(self)
        self.settings_bar.list._page = self.settings_page
        self.settings_page.hide()
        self.server_hover = ServerHover(self)

        self.grid_lo.addWidget(self.menu_bar, 0, 0)
        self.grid_lo.addWidget(self.settings_bar, 0, 0)
        self.grid_lo.addWidget(self.wizard, 0, 1, QtCore.Qt.AlignCenter)
        self.grid_lo.addWidget(self.login, 0, 1, QtCore.Qt.AlignCenter)
        self.grid_lo.addWidget(self.projects, 0, 1)
        self.grid_lo.addWidget(self.settings_page, 0, 1)
        self.page_refresh('projects', force_update=True, init=True)
    
    def check_wizard(self, force_update=False):
        self.menu_bar.servers.refresh(force_update)
        if self.menu_bar.servers.count():
            return False

        print('INFO: Starting Wizard')

        for i in reversed(range(self.grid_lo.count())):
            widget = self.grid_lo.itemAt(i).widget()
            if (
                widget.objectName() == 'NotificationHeader' or 
                widget.objectName() == 'PopUpDialog' or 
                widget.objectName() == 'ServerHover'
            ):
                widget.deleteLater()
        
        self.overlay_enabled = False
        self.menu_bar.hide()
        self.wizard.build()
        self.wizard.show()
        self.login.hide()
        self.projects.hide()
        self.settings_bar.hide()
        self.settings_page.hide()
        return True

    def page_refresh(self, redirect, force_update=False, init=False, logged=False):
        if self.check_wizard(force_update):
            return

        if not self.overlay_enabled:
            self.notification_header = NotificationHeader(self)
            self.popup = PopUpDialog(self)
            self.grid_lo.addWidget(self.server_hover, 0, 0, 0, 2, QtCore.Qt.AlignTop)
            self.grid_lo.addWidget(self.notification_header, 0, 0, 0, 2, QtCore.Qt.AlignTop)
            self.grid_lo.addWidget(self.popup, 0, 0, 0, 2)
            self.overlay_enabled = True

        if redirect == 'projects':
            self.menu_bar.show()
            self.wizard.hide()
            self.settings_bar.hide()
            self.settings_page.hide()

            if init:
                self._ctrl.set_current_host()
                self.menu_bar.servers.refresh()

            if logged or self._ctrl.check_log_in():
                self.login.hide()
                self.projects.refresh()
                self.projects.show()
                if init:
                    self.notification_header.pop('Connected', 'valid', True)
            else:
                self.login.build()
                self.login.show()
                self.projects.hide()
                if self.menu_bar.servers.current_item:
                    self.menu_bar.servers.current_item.refresh()
            
            return
        
        if redirect == 'settings':
            self.settings_bar.list.refresh(True)
            if self.settings_bar.list.current_item.text() != 'Global':
                self.settings_bar.list.setSelected(self.settings_bar.list.layout.itemAt(0).widget())

            self.menu_bar.hide()
            self.login.hide()
            self.projects.hide()
            self.settings_bar.show()
            self.settings_page.refresh()
            self.settings_page.show()
            return

    def mousePressEvent(self, event):
        focused_widget = QtWidgets.QApplication.focusWidget()
        if isinstance(focused_widget, LineEdit) or isinstance(focused_widget, ComboBox):
            focused_widget.clearFocus()
        QtWidgets.QMainWindow.mousePressEvent(self, event)


class PopUpDialog(QtWidgets.QFrame):

    ICONS = {
        "delete": ('icons.gui', 'trash_normal')
    }
    DESCRIPTION = {
        "delete": "You're sure you want to delete this {item_type}"
    }

    def __init__(self, mw):
        QtWidgets.QFrame.__init__(self)
        self.setObjectName('PopUpDialog')

        self._mw = mw
        self._ctrl = mw._ctrl
        self._signal = None

        self.hide()
        self.build()

    def build(self):
        self.container_lo = QtWidgets.QVBoxLayout(self)
        
        self.message_widget = QtWidgets.QFrame()
        self.message_widget.setObjectName('PopUpMessage')
        self.message_widget.setFixedWidth(300)
        self.message_lo = QtWidgets.QVBoxLayout(self.message_widget)
        self.message_lo.setSpacing(10)

        self.icon_lo = QtWidgets.QVBoxLayout()
        self.icon_lo.setContentsMargins(20,20,20,20)
        self.message_lo.addLayout(self.icon_lo)

        self.icon_lbl = LabelIcon()
        self.icon_lo.addWidget(self.icon_lbl)

        self.main_label = QtWidgets.QLabel()
        self.main_label.setWordWrap(True)
        self.main_label.setAlignment(QtCore.Qt.AlignCenter)
        self.main_label.setObjectName('PopUpMainLabel')
        self.description_label = QtWidgets.QLabel()
        self.description_label.setWordWrap(True)
        self.description_label.setAlignment(QtCore.Qt.AlignCenter)

        self.message_lo.addWidget(self.main_label)
        self.message_lo.addWidget(self.description_label)

        self.buttons_lo = QtWidgets.QHBoxLayout()
        self.buttons_lo.setContentsMargins(10,10,10,20)
        self.message_lo.addLayout(self.buttons_lo)

        self.secondary_btn = QtWidgets.QPushButton('Cancel')
        self.secondary_btn.setObjectName('PopUpButton')
        self.secondary_btn.clicked.connect(self._on_secondary_button_clicked)
        self.buttons_lo.addWidget(self.secondary_btn)

        self.primary_btn = QtWidgets.QPushButton('')
        self.primary_btn.setObjectName('PopUpButton')
        self.primary_btn.clicked.connect(self._on_primary_button_clicked)
        self.buttons_lo.addWidget(self.primary_btn)

        self.container_lo.addWidget(self.message_widget, alignment=QtCore.Qt.AlignCenter)

    def pop(self, action, p_type=None, value=None, signal=None):
        self._mw.popup.show()
        self._type = p_type
        self._signal = signal
        
        if action == 'delete':
            self.icon_lbl.setIcon(self.ICONS.get(action), 50)
            self.icon_lbl.setAlignment(QtCore.Qt.AlignCenter)
            self.main_label.setText(f'{action.capitalize()} {value}?')
            self.description_label.setText(self.DESCRIPTION.get(action).format(item_type=self._type))
            self.primary_btn.setText(action.capitalize())
            self.primary_btn.setProperty(action, True)
            self.primary_btn.style().polish(self.primary_btn)
            return
        
        if action == 'updating' or action == 'launching':
            self.icon_lbl.hide()
            self.main_label.setText(f'{action.capitalize()} Libreflow')
            self.description_label.setText('Please wait...')
            self.secondary_btn.hide()
            self.primary_btn.hide()
            return

    def _on_secondary_button_clicked(self):
        self.hide()

    def _on_primary_button_clicked(self):
        self._signal()
        self.hide()


class NotificationHeader(QtWidgets.QFrame):

    def __init__(self, mw):
        QtWidgets.QFrame.__init__(self)
        self.setObjectName('NotificationHeader')

        self._mw = mw
        self._ctrl = mw._ctrl

        self.hide()

        self.setAttribute(QtCore.Qt.WA_TransparentForMouseEvents)
      
        container_lo = QtWidgets.QVBoxLayout(self)
        
        self.message_widget = QtWidgets.QFrame()
        message_lo = QtWidgets.QHBoxLayout(self.message_widget)

        self.message_widget.setObjectName('NotificationMessage')
        self.message_widget.setFixedWidth(400)

        self.label = QtWidgets.QLabel('')
        self.label.setWordWrap(True)
        self.label.setAlignment(QtCore.Qt.AlignCenter)
        message_lo.addWidget(self.label)

        container_lo.addWidget(self.message_widget, 0, QtCore.Qt.AlignCenter)

        self.timer = QtCore.QTimer()
        self.timer.timeout.connect(self._hide_message)

    def pop(self, text, n_type, temporary=False):
        self._mw.notification_header.show()
        self._type = n_type
        self.label.setText(text)
        self.message_widget.setProperty(self._type, True)
        self.message_widget.style().polish(self.message_widget)
        if temporary:
            self.timer.start(4000)
    
    def _hide_message(self):
        self.message_widget.setProperty(self._type, False)
        self._mw.notification_header.hide()
        self.timer.stop()


class ServerHover(QtWidgets.QFrame):

    def __init__(self, mw):
        QtWidgets.QFrame.__init__(self)
        self.setObjectName('ServerHover')
        self.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)

        self._mw = mw
        self._ctrl = mw._ctrl

        self.ref_geometry = None
        self.width = None
        self.height = None

        self.hide()
        self.build()

    def build(self):
        self.container_lo = QtWidgets.QVBoxLayout(self)
        
        self.box_widget = QtWidgets.QWidget()
        self.box_widget.setObjectName('BoxWidget')
        self.box_widget.setFixedHeight(57)
        self.box_widget.installEventFilter(self)
        box_lo = QtWidgets.QHBoxLayout(self.box_widget)
        box_lo.setContentsMargins(0,0,0,0)

        horizontal_spacer = QtWidgets.QSpacerItem(
            self._mw.menu_bar.width(), 0, 
            QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed
        )
        box_lo.addItem(horizontal_spacer)

        self.box_frame = QtWidgets.QFrame()
        box_frame_lo = QtWidgets.QVBoxLayout(self.box_frame)
        box_lo.addWidget(self.box_frame)

        self.box_frame.setStyleSheet('background-color: #f2eff1; border-radius: 5px;')
        self.box_frame.setFixedWidth(200)
        self.box_frame.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)

        server_lo = QtWidgets.QHBoxLayout()
        box_frame_lo.addLayout(server_lo)

        self.server_icon = LabelIcon(('icons.gui', 'server'))
        self.server_label = QtWidgets.QLabel('')
        self.server_label.setWordWrap(True)
        # server_lo.addWidget(self.server_icon)
        server_lo.addWidget(self.server_label)
        server_lo.addStretch()

        user_lo = QtWidgets.QHBoxLayout()
        box_frame_lo.addLayout(user_lo)

        self.user_icon = LabelIcon(('icons.gui', 'user'))
        self.user_label = QtWidgets.QLabel('')
        self.user_label.setWordWrap(True)
        self.user_log_out = QtWidgets.QToolButton()
        self.user_log_out.setObjectName('ServerLogOutButton')
        self.user_log_out.clicked.connect(self._on_log_out_button_clicked)
        # user_lo.addWidget(self.user_icon)
        user_lo.addWidget(self.user_label)
        user_lo.addStretch()
        user_lo.addWidget(self.user_log_out)

        self.container_lo.addWidget(self.box_widget)

    def settingMask(self):
        self.region = QtGui.QRegion(self.frameGeometry())
        coord = QtCore.QPoint(
            self._mw.menu_bar.layout().getContentsMargins()[0]+self.width,
            self._mw.menu_bar.layout().getContentsMargins()[1]+self.height
        )
        self.ref_geometry.moveTopLeft(coord)
        self.region -= QtGui.QRegion(self.ref_geometry, QtGui.QRegion.Ellipse)
        self.setMask(self.region)

    def pop(self, server):
        self.ref_geometry = server.circle.frameGeometry()
        self.width = server.circle.frameGeometry().x()
        self.height = server.frameGeometry().y()

        self.container_lo.setContentsMargins(9,9 + self.height,9,9)
        self.server_label.setText(server._data['display_name'])
        if server._current_user:
            self.user_label.setText(f'{server._current_user} ({os.environ["LF_LAUNCHER_SITE_NAME"]})')
            self.user_label.show()
            self.user_log_out.show()
        else:
            self.user_label.hide()
            self.user_log_out.hide()
        self.settingMask()
        self.show()

    def resizeEvent(self, event):
        if self.ref_geometry:
            self.settingMask()

    def eventFilter(self, obj, event):
        if obj.objectName() == 'BoxWidget':
            if event.type() == QtCore.QEvent.Enter:
                return True
            elif event.type() == QtCore.QEvent.Leave:
                self.hide()
                return True
        return False

    def _on_log_out_button_clicked(self):
        self._ctrl.log_out()
        self._mw.page_refresh('projects', True)
        self.hide()


class WizardPage(QtWidgets.QFrame):

    def __init__(self, mw):
        QtWidgets.QFrame.__init__(self)
        self._mw = mw
        self._ctrl = mw._ctrl

        self.container_lo = QtWidgets.QVBoxLayout(self)
        
    def build(self):
        if self.container_lo.count() > 0:
            for i in reversed(range(self.container_lo.count())):
                self.container_lo.itemAt(i).widget().deleteLater()
        
        self.landing = self.landing_screen()
        self.setup = self.server_setup()
        self.container_lo.addWidget(self.landing)
        self.container_lo.addWidget(self.setup)

    def landing_screen(self):
        landing_widget = QtWidgets.QWidget()
        landing_lo = QtWidgets.QVBoxLayout(landing_widget)

        main_text = QtWidgets.QLabel('Welcome to Libreflow Launcher!')
        main_text.setStyleSheet('font-weight: bold; font-size: 36px;')
        landing_lo.addWidget(main_text, 0, QtCore.Qt.AlignCenter)

        start_button = QtWidgets.QPushButton('Let\'s add a server')
        start_button.setAutoDefault(True)
        start_button.setObjectName('LoginButton')
        start_button.clicked.connect(self._on_start_button_clicked)
        landing_lo.addWidget(start_button, 0, QtCore.Qt.AlignCenter)

        return landing_widget

    def server_setup(self):
        server_widget = QtWidgets.QWidget()
        server_widget.hide()
        self.server_lo = QtWidgets.QVBoxLayout(server_widget)

        main_text = QtWidgets.QLabel('Enter server parameters')
        main_text.setStyleSheet('font-weight: bold; font-size: 36px;')
        self.server_lo.addWidget(main_text, 0, QtCore.Qt.AlignCenter)

        self.name_input = LineEdit(options=['mandatory'])
        self.name_input.setObjectName('ServerInput')
        self.name_input.setPlaceholderText('Server name')
        self.server_lo.addWidget(self.name_input, 0, QtCore.Qt.AlignCenter)
        
        self.host_input = LineEdit(options=['mandatory', 'host'])
        self.host_input.setObjectName('ServerInput')
        self.host_input.setPlaceholderText('Host')
        self.server_lo.addWidget(self.host_input, 0, QtCore.Qt.AlignCenter)

        add_button = QtWidgets.QPushButton('Add')
        add_button.setObjectName('LoginButton')
        add_button.setDefault(True)
        add_button.clicked.connect(self._on_add_button_clicked)
        self.server_lo.addWidget(add_button, 0, QtCore.Qt.AlignCenter)

        return server_widget

    def _on_start_button_clicked(self):
        self.landing.hide()
        self.setup.show()

    def _on_add_button_clicked(self):
        for i in range(self.server_lo.count()):
            widget = self.server_lo.itemAt(i).widget()
            if widget.property('warning'):
                return

        api_host = self.host_input.text()
        api_port = '5500'
        
        if ':' in self.host_input.text():
            api_host, api_port = self.host_input.text().split(':')

        data = dict(
            display_name=self.name_input.text(),
            api_host=api_host,
            api_port=api_port
        )
        uid = self._ctrl.update_server(data)
        self._mw.page_refresh('projects', True)


class SettingsPage(QtWidgets.QFrame):

    def __init__(self, mw):
        QtWidgets.QFrame.__init__(self)
        self._mw = mw
        self._sb = mw.settings_bar
        self._ctrl = mw._ctrl

        self.page_lo = QtWidgets.QVBoxLayout(self)

        self.build()

    def build(self):
        header = QtWidgets.QHBoxLayout()
        self.page_lo.addLayout(header)
        
        self.page_text = QtWidgets.QLabel('')
        self.page_text.setStyleSheet('font-weight: bold; font-size: 24px;')
        header.addWidget(self.page_text)

        self.delete_btn = QtWidgets.QToolButton()
        self.delete_btn.setObjectName("SettingsDeleteButton")
        self.delete_btn.clicked.connect(self._on_delete_button_clicked)
        header.addWidget(self.delete_btn)
        
        header.addStretch()
        
        close_btn = QtWidgets.QPushButton(
            resources.get_icon(('icons.gui', 'close'), color=QtGui.QColor('#26262f')), ''
        )
        close_btn.setObjectName('MenuBarButton')
        close_btn.clicked.connect(self._on_close_button_clicked)
        header.addWidget(close_btn)

        self.content = QtWidgets.QGridLayout()
        # Left, top, right, bottom
        self.content.setContentsMargins(0, 20, 0, 20)
        self.page_lo.addLayout(self.content)

        self.stretch_widget = QtWidgets.QWidget()
        stretch_lo = QtWidgets.QVBoxLayout(self.stretch_widget)
        self.page_lo.addWidget(self.stretch_widget)

        self.footer_empty = QtWidgets.QWidget()
        self.footer_empty.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        self.footer_empty_lo = QtWidgets.QHBoxLayout(self.footer_empty)
        self.footer_empty_lo.setContentsMargins(0,0,0,0)
        self.page_lo.addWidget(self.footer_empty)
        
        spacer = QtWidgets.QSpacerItem(0, 27, QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        self.footer_empty_lo.addItem(spacer)
        
        self.footer = QtWidgets.QWidget()
        self.footer.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        footer_lo = QtWidgets.QHBoxLayout(self.footer)
        footer_lo.setContentsMargins(0,0,0,0)
        self.page_lo.addWidget(self.footer)
        self.footer.hide()

        cancel_btn = QtWidgets.QPushButton('Cancel')
        cancel_btn.setObjectName('SettingsFooterButton')
        cancel_btn.clicked.connect(self._on_cancel_button_clicked)
        self.save_btn = QtWidgets.QPushButton('Save')
        self.save_btn.setObjectName("SettingsFooterButton")
        self.save_btn.clicked.connect(self._on_save_button_clicked)

        footer_lo.addStretch()
        footer_lo.addWidget(cancel_btn)
        footer_lo.addWidget(self.save_btn)
    
    def refresh(self):
        self.current_item = self._sb.list.current_item
        if self.current_item.elided:
            self.page_text.setText(self.current_item.full_text)
        else:
            self.page_text.setText(self.current_item.text())

        if self.content.count() > 0:
            for i in reversed(range(self.content.count())):
                self.content.itemAt(i).widget().deleteLater()

        for i, group in enumerate(self.current_item.get_content()):
            for x, widget in enumerate(group):
                self.content.addWidget(widget, i, x)
        
        if type(self.current_item._data) == dict:
            self.delete_btn.show()
        else:
            self.delete_btn.hide()
        
        if self.current_item._data == "About":
            self.page_text.setText('')
            self.stretch_widget.hide()
        else:
            self.stretch_widget.show()

        self.footer_empty.show()
        self.footer.hide()

    def toggle_footer(self):
        for i in range(self.content.count()):
            widget = self.content.itemAt(i).widget()
            if widget.property('edited'):
                self.footer_empty.hide()
                self.footer.show()
                return
        
        self.footer_empty.show()
        self.footer.hide()

    def _on_close_button_clicked(self):
        self._mw.page_refresh('projects', True)

    def _on_delete_button_clicked(self):
        if self._mw.popup.isVisible():
            self._ctrl.remove_server(self.current_item._data['uid'])
            self._mw.menu_bar.servers.refresh(True)
            self._mw.menu_bar.servers.orderChanged.emit(self._mw.menu_bar.servers.get_order_data())
            self._sb.list.refresh(True)
            self._mw.settings_page.refresh()
            return self._mw.notification_header.pop('Server deleted', 'valid', True)

        self._mw.popup.pop('delete', 'server', self.page_text.text(), self._on_delete_button_clicked)

    def _on_cancel_button_clicked(self):
        self.refresh()

    def _on_save_button_clicked(self):
        data = {}
        for i in range(self.content.count()):
            widget = self.content.itemAt(i).widget()
            if widget.property('warning'):
                return self._mw.notification_header.pop("One or more inputs are incorrect", 'error', True)

            if widget.property('edited'):
                if isinstance(widget, LineEdit):
                    if widget.config_key == 'api_host':
                        data[widget.config_key] = widget.text().split(':')[0]
                        data['api_port'] = widget.text().split(':')[1]
                    else:
                        data[widget.config_key] = widget.text()
                if isinstance(widget, ComboBox):
                    data[widget.config_key] = widget.currentData()
                    if widget.config_key == 'python_path':
                        data['python_version'] = widget.currentText()
                widget.reset()
        
        focused_widget = QtWidgets.QApplication.focusWidget()
        focused_widget.clearFocus()
        
        if type(self.current_item._data) == dict:
            self._ctrl.update_server(data, self.current_item._data['uid'])
            self._ctrl.set_current_host()
            self.current_item.refresh()
            self.refresh()
        else:
            self._ctrl.update_settings(data)
        
        self._mw.notification_header.pop("Saved", 'valid', True)
        self.toggle_footer()


class SettingsContent(QtWidgets.QPushButton):

    def __init__(self, settings_list, data):
        super(SettingsContent, self).__init__('')
        self.setObjectName('SettingsButton')
        self.setProperty('selected', False)

        self.settings_list = settings_list
        self._data = data
        self._mw = settings_list._mw
        self._ctrl = settings_list._ctrl

        self.elided = False
        self.full_text = None
             
        self.refresh()

    def refresh(self):
        if type(self._data) == str:
            self.setText(self._data)
        else:
            self.setText(self._data['display_name'])

        if len(self.text()) > 20:
            self.elided = True
            self.full_text = self.text()
            self.setText(''.join([x[0].upper() for x in self.text().split(' ')]))
    
    def setSelected(self):
        if not self.property('selected'):
            self.setProperty('selected', True)
            self.style().polish(self)
        else:
            self.setProperty('selected', False)
            self.style().polish(self)

    def mouseReleaseEvent(self, event):
        if event.button() == QtCore.Qt.LeftButton:
            if self.settings_list.current_item == self:
                return
            
            self.settings_list.setSelected(self)


class GlobalSettings(SettingsContent):

    def get_content(self):
        contents = []

        data = self._ctrl.fetch_settings()

        install_label = QtWidgets.QLabel("Installation directory")
        install_input = LineEdit(
            data['install_dir'],
            ['mandatory', 'path'],
            'install_dir',
            self.settings_list
        )
        install_input.setObjectName("SettingsInput")
        first_group = [install_label, install_input]
        contents.append(first_group)

        python_label = QtWidgets.QLabel("Default python version")
        python_input = ComboBox(
            config_key='python_path',
            parent=self.settings_list
        )
        python_input.setObjectName("SettingsInput")
        python_input.view().window().setWindowFlags(QtCore.Qt.Popup | QtCore.Qt.FramelessWindowHint | QtCore.Qt.NoDropShadowWindowHint)
        python_input.view().window().setAttribute(QtCore.Qt.WA_TranslucentBackground)
        
        python_versions = self._mw._ctrl.fetch_python_versions()
        for version_data in python_versions:
            python_input.addItem(version_data['version'], version_data['path'])
        
        python_input_index = python_input.findData(data['python_path'])
        python_input.setCurrentIndex(python_input_index)

        second_group = [python_label, python_input]
        contents.append(second_group)

        python_input.enable_check()

        return contents


class ServerSettings(SettingsContent):

    def get_content(self):
        contents = []

        name_label = QtWidgets.QLabel("Host name")
        name_input = LineEdit(
            self._data['display_name'],
            ['mandatory'],
            'display_name',
            self.settings_list
        )
        name_input.setObjectName("SettingsInput")
        first_group = [name_label, name_input]
        contents.append(first_group)

        host_label = QtWidgets.QLabel("Host address")
        host_input = LineEdit(
            f"{self._data['api_host']}:{self._data['api_port']}",
            ['mandatory', 'host'],
            'api_host',
            self.settings_list
        )
        host_input.setObjectName("SettingsInput")
        second_group = [host_label, host_input]
        contents.append(second_group)

        return contents


class About(SettingsContent):

    def get_content(self):
        contents = []

        about_widget = QtWidgets.QWidget()
        
        about_lo = QtWidgets.QVBoxLayout(about_widget)
        about_lo.setContentsMargins(0,0,0,0)

        about_lo.addStretch()

        logo = LabelIcon(('icons.gui', 'feespeciales_icon'), 100)
        about_lo.addWidget(logo, 0, QtCore.Qt.AlignCenter)

        software_text = QtWidgets.QLabel('Libreflow Launcher')
        software_text.setStyleSheet('font-weight: bold; font-size: 36px;')
        about_lo.addWidget(software_text, 0, QtCore.Qt.AlignCenter)

        version_text = QtWidgets.QLabel('Version 1.0.0')
        about_lo.addWidget(version_text, 0, QtCore.Qt.AlignCenter)

        web_buttons = QtWidgets.QWidget()
        web_buttons_lo = QtWidgets.QHBoxLayout(web_buttons)
        
        gitlab_btn = QtWidgets.QPushButton(
            resources.get_icon(('icons.gui', 'gitlab')), 'GitLab'
        )
        gitlab_btn.setIconSize(QtCore.QSize(30,30))
        gitlab_btn.setObjectName("GitLabButton")
        gitlab_btn.clicked.connect(self._on_gitlab_button_clicked)
        web_buttons_lo.addWidget(gitlab_btn)

        pypi_btn = QtWidgets.QPushButton(
            resources.get_icon(('icons.gui', 'pypi')), 'PyPI'
        )
        pypi_btn.setIconSize(QtCore.QSize(30,30))
        pypi_btn.setObjectName("PyPIButton")
        pypi_btn.clicked.connect(self._on_pypi_button_clicked)
        web_buttons_lo.addWidget(pypi_btn)

        about_lo.addWidget(web_buttons, 0, QtCore.Qt.AlignCenter)
        about_lo.addStretch()

        contents.append([about_widget])

        return contents
    
    def _on_gitlab_button_clicked(self):
        webbrowser.open("https://gitlab.com/lfs.coop/libreflow/libreflow_launcher")

    def _on_pypi_button_clicked(self):
        webbrowser.open("https://gitlab.com/lfs.coop")


class SettingsList(QtWidgets.QScrollArea):

    def __init__(self, sb):
        super(SettingsList, self).__init__()
        self._sb = sb
        self._mw = sb._mw
        self._ctrl = sb._ctrl
        self._page = None

        self.current_item = None

        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setWidgetResizable(True)

        self.setStyleSheet(
            '''
            QScrollArea {
                background-color: transparent;
                border: none;
            }
            '''
        )

    def refresh(self, force_update=False):
        if force_update:
            self.current_item = None

            container = QtWidgets.QWidget()
            container.setObjectName('ScrollAreaContainer')
            container.setStyleSheet(
                '''
                #ScrollAreaContainer {
                    background-color: transparent;
                    border: none;
                }
                '''
            )
            self.layout = QtWidgets.QVBoxLayout(container)
            self.layout.setAlignment(QtCore.Qt.AlignTop)
            # self.layout.setSpacing(30)

            global_btn = GlobalSettings(self, 'Global')
            self.layout.addWidget(global_btn)
        
            for i in range(self._mw.menu_bar.servers.layout.count()):
                widget = self._mw.menu_bar.servers.layout.itemAt(i).widget()
                if widget:
                    btn = ServerSettings(self, widget._data)
                    self.layout.addWidget(btn)
            
            about_btn = About(self, 'About')
            self.layout.addWidget(about_btn)

            self.layout.addStretch(1)

            # Left, top, right, bottom
            self.layout.setContentsMargins(0,0,0,0)
            self.setWidget(container)

        self.setSelected()

    def count(self):
        return self.layout.count()-1

    def setSelected(self, item=None):
        if not self.current_item:
            for i in range(self.layout.count()):
                item = self.layout.itemAt(i).widget()
                if item:
                    if item.text() == "Global":
                        item.setSelected()
                        self.current_item = item
                        break
            return

        self.current_item.setSelected()
        self.current_item = item
        self.current_item.setSelected()
        self._mw.settings_page.refresh()


class SettingsBar(QtWidgets.QFrame):

    def __init__(self, mw):
        QtWidgets.QFrame.__init__(self)
        self._mw = mw
        self._ctrl = mw._ctrl

        self.setFixedWidth(175)
        self.setObjectName('MenuBar')
        self.setStyleSheet('#MenuBar { background-color: #f2eff1; }')
       
        vlo = QtWidgets.QVBoxLayout(self)
        # vlo.setContentsMargins(0,0,0,0)

        self.list = SettingsList(self)
        vlo.addWidget(self.list)
        vlo.addStretch()


class ServerItem(QtWidgets.QWidget):

    def __init__(self, servers_list, data, current_user):
        super(ServerItem, self).__init__()
        self.setFixedHeight(45)

        self.servers_list = servers_list
        self._data = data
        self._current_user = current_user
        self._mw = servers_list._mw
        self._ctrl = servers_list._ctrl

        self.setWindowFlags(QtCore.Qt.FramelessWindowHint)
        self.setAttribute(QtCore.Qt.WA_TranslucentBackground)
            
        # shadow = QtWidgets.QGraphicsDropShadowEffect(self)
        # shadow.setBlurRadius(40)
        # shadow.setXOffset(0)
        # shadow.setYOffset(5)
        # shadow.setColor(QtGui.QColor(0, 0, 0, 81))
        # self.setGraphicsEffect(shadow)
        
        self.build()

    def build(self):
        container = QtWidgets.QGridLayout(self)
        container.setContentsMargins(0,0,0,0)

        self.circle = QtWidgets.QWidget()
        self.circle.setFixedSize(45,45)
        self.circle.setObjectName('ServerItem')
        self.circle.setProperty('selected', False)
        self.circle.installEventFilter(self)
        circle_lo = QtWidgets.QVBoxLayout(self.circle)
        circle_lo.setContentsMargins(0,0,0,0)

        self.text = QtWidgets.QLabel(self._data['code'])
        self.text.setObjectName('ServerItem')
        self.text.setProperty('selected', False)
        self.text.setAlignment(QtCore.Qt.AlignVCenter)
        circle_lo.addWidget(self.text, alignment=QtCore.Qt.AlignCenter)

        container.addWidget(self.circle, 0, 0, alignment=QtCore.Qt.AlignCenter)

        self.arrow = LabelIcon(('icons.gui', 'arrow-right'), 7, QtGui.QColor('white'))
        self.arrow.setAlignment(QtCore.Qt.AlignCenter)
        self.arrow.setFixedSize(15,15)
        self.arrow.setStyleSheet('border-radius: 7px; background-color: #5553ac;')
        self.arrow.installEventFilter(self)
        container.addWidget(self.arrow, 0, 0, alignment=QtCore.Qt.AlignRight)
        self.arrow.hide()

    def refresh(self):
        self._current_user = self._ctrl.fetch_user_name(self._data['uid'])

    def setSelected(self):
        if not self.circle.property('selected'):
            self.circle.setProperty('selected', True)
            self.text.setProperty('selected', True)
            self.arrow.show()
        else:
            self.circle.setProperty('selected', False)
            self.text.setProperty('selected', False)
            self.arrow.hide()
        self.circle.style().polish(self.circle)
        self.text.style().polish(self.text)

    def eventFilter(self, obj, event):
        if event.type() == QtCore.QEvent.Enter:
            if not self._mw.server_hover.isVisible():
                self._mw.server_hover.pop(self)
            return True
        return False

    def mouseMoveEvent(self, event):
        if event.buttons() == QtCore.Qt.LeftButton:
            self._mw.server_hover.hide()

            drag = QtGui.QDrag(self)
            mime = QtCore.QMimeData()
            drag.setMimeData(mime)
            drag.setHotSpot(QtCore.QPoint(25, 25))
            
            pixmap = QtGui.QPixmap(self.size())
            pixmap.fill(QtCore.Qt.transparent)
            self.render(pixmap)
            drag.setPixmap(pixmap)

            store_status = self.circle.property('selected')
            self.circle.setProperty('selected', False)
            self.circle.setProperty('drag', True)
            self.text.setProperty('selected', False)
            self.circle.style().polish(self.circle)
            self.text.style().polish(self.text)
            self.text.hide()
            self.arrow.hide()
            
            drag.exec_(QtCore.Qt.MoveAction)

            self.text.show()
            self.circle.setProperty('drag', False)
            self.circle.style().polish(self.circle)
            if store_status:
                self.setSelected()

    def mouseReleaseEvent(self, event):
        if event.button() == QtCore.Qt.LeftButton:
            if self.servers_list._ctrl.server_uid == self._data['uid']:
                return
            
            self.servers_list._ctrl.set_current_host(self._data['uid'])
            self._mw.page_refresh('projects')


class ServersList(QtWidgets.QScrollArea):

    orderChanged = QtCore.Signal(list)

    def __init__(self, mb):
        super(ServersList, self).__init__()
        self._mb = mb
        self._mw = mb._mw
        self._ctrl = mb._ctrl

        self.current_item = None

        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setWidgetResizable(True)
        self.setAcceptDrops(True)

        self.setStyleSheet(
            '''
            QScrollArea {
                background-color: transparent;
                border: none;
            }
            QScrollBar {
                height: 0px;
            }
            '''
        )

    def refresh(self, force_update=False):
        if force_update:
            self.current_item = None
            servers = self._ctrl.fetch_servers()

            container = QtWidgets.QWidget()
            container.setObjectName('ScrollAreaContainer')
            container.setStyleSheet(
                '''
                #ScrollAreaContainer {
                    background-color: transparent;
                    border: none;
                }
                '''
            )
            self.layout = QtWidgets.QVBoxLayout(container)
            self.layout.setAlignment(QtCore.Qt.AlignTop)
        
            for server in servers:
                current_user = self._ctrl.fetch_user_name(server['uid'])
                item = ServerItem(self, server, current_user)
                self.layout.addWidget(item)
            
            # Left, top, right, bottom
            self.layout.setContentsMargins(0,0,0,0)
            self.setWidget(container)

            # Auto set if there is only one server
            if len(servers) == 1 and self._ctrl.server_uid == None:
                self._ctrl.set_current_host(server['uid'])

        self.setSelected()

    def count(self):
        return self.layout.count()

    def setSelected(self):
        # Unselect
        if self.current_item:
            index = self.layout.indexOf(self.current_item)
            self.layout.itemAt(index).widget().setSelected()

        # Select
        current_host = self._ctrl.server_uid
        for i in range(self.layout.count()):
            item = self.layout.itemAt(i).widget()
            if isinstance(item, ServerItem):
                if item._data['uid'] == current_host:
                    item.setSelected()
                    self.current_item = item
                    break

    def dragEnterEvent(self, event):
        event.accept()

    def dropEvent(self, event):
        pos = event.pos()
        source_widget = event.source()

        # Check if we don't drop on the same widget
        if not source_widget.geometry().contains(pos):
            target_widget = None
            for i in range(self.layout.count()):
                w = self.layout.itemAt(i).widget()
                if w.geometry().contains(pos):
                    target_widget = w
            if target_widget is None:
                # For let switch if we drop at the bottom of the list
                if not pos.y() > w.y():
                    return
            
            self.layout.insertWidget(self.layout.indexOf(target_widget), source_widget)
            self.orderChanged.emit(self.get_order_data())

        event.accept()

    def get_order_data(self):
        data = []
        for i in range(self.layout.count()):
            item = self.layout.itemAt(i).widget()
            data.append(item._data['uid'])
        
        self._ctrl.update_servers_order(data)
        return data


class MenuBar(QtWidgets.QFrame):

    def __init__(self, mw):
        QtWidgets.QFrame.__init__(self)
        self._mw = mw
        self._ctrl = mw._ctrl

        self.setFixedWidth(75)
        self.setObjectName('MenuBar')
        self.setStyleSheet('#MenuBar { background-color: #f2eff1; }')
       
        vlo = QtWidgets.QVBoxLayout(self)
        vlo.setContentsMargins(9,15,9,15)

        self.servers = ServersList(self)
        vlo.addWidget(self.servers)
        
        settings_btn = QtWidgets.QPushButton(
            resources.get_icon(('icons.gui', 'settings_outline'), color=QtGui.QColor('#26262f')), ''
        )
        settings_btn.setObjectName('MenuBarButton')
        settings_btn.clicked.connect(self._on_settings_button_clicked)
        vlo.addWidget(settings_btn, alignment=QtCore.Qt.AlignCenter)
       
    def _on_settings_button_clicked(self):
        self._mw.page_refresh('settings')


class LoginPage(QtWidgets.QFrame):

    def __init__(self, mw):
        QtWidgets.QFrame.__init__(self)
        self._mw = mw
        self._ctrl = mw._ctrl
       
        self.login_lo = QtWidgets.QVBoxLayout(self)

    def build(self):
        if self.login_lo.count() > 0:
            for i in reversed(range(self.login_lo.count())):
                self.login_lo.itemAt(i).widget().deleteLater()

        logo = LabelIcon(('icons.gui', 'feespeciales_icon'), 100)
        self.login_lo.addWidget(logo, 0, QtCore.Qt.AlignCenter)

        self.login_input = LineEdit()
        self.login_input.setObjectName('LoginInput')
        self.login_input.setPlaceholderText('Login')
        self.login_lo.addWidget(self.login_input, 0, QtCore.Qt.AlignCenter)

        self.password_input = LineEdit()
        self.password_input.setObjectName('LoginInput')
        self.password_input.setPlaceholderText('Password')
        self.password_input.setEchoMode(QtWidgets.QLineEdit.Password)
        self.login_lo.addWidget(self.password_input, 0, QtCore.Qt.AlignCenter)

        connect_button = QtWidgets.QPushButton('Connect')
        connect_button.setObjectName('LoginButton')
        connect_button.setDefault(True)
        connect_button.clicked.connect(self._on_connect_button_clicked)
        self.login_lo.addWidget(connect_button, 0, QtCore.Qt.AlignCenter)

    def _on_connect_button_clicked(self):
        logged_in = self._ctrl.log_in(self.login_input.text(), self.password_input.text())
        if logged_in:
            self._mw.page_refresh('projects', True, logged=True)
            self._mw.notification_header.pop('Connected', 'valid', True)
        else:
            self._mw.notification_header.pop('Authentification invalid', 'error', True)


class ProjectItem(QtWidgets.QWidget):

    def __init__(self, projects_list, data):
        super(ProjectItem, self).__init__()
        self.setObjectName('ProjectItem')

        self.projects_list = projects_list
        self._mw = projects_list._mw
        self._ctrl = projects_list._ctrl

        self._status = data['updated']
        self._data = data['data']
        self._uid = self._data['uid']
            
        self.setAttribute(QtCore.Qt.WA_StyledBackground, True)

        ### For grid view
        # self.setMaximumWidth(335)
        # self.setMaximumHeight(200)

        shadow = QtWidgets.QGraphicsDropShadowEffect(self)
        shadow.setBlurRadius(25)
        shadow.setXOffset(0)
        shadow.setYOffset(5)
        shadow.setColor(QtGui.QColor(0, 0, 0, 81))
        self.setGraphicsEffect(shadow)
        
        self.setStyleSheet(
            '''
            QLabel {
                color: white;
            }
            #ProjectItem {
                background-color: #2f2e5f;
                border-radius: 10px;
            }
            '''
        )

        self.build()
        self.refresh()

    def build(self):
        container = QtWidgets.QVBoxLayout(self)
        container.setContentsMargins(20,10,20,10)

        footer = QtWidgets.QWidget()
        footer_lo = QtWidgets.QHBoxLayout(footer)
        footer_lo.setContentsMargins(0,0,0,0)
        container.addWidget(footer, alignment=QtCore.Qt.AlignBottom)

        self.status_lbl = QtWidgets.QLabel('')
        footer_lo.addWidget(self.status_lbl, alignment=QtCore.Qt.AlignVCenter)

        self.project_name = QtWidgets.QLabel()
        self.project_name.setStyleSheet('color: white; font-size: 24px;')

        footer_lo.addWidget(self.project_name, alignment=QtCore.Qt.AlignVCenter)
        footer_lo.addStretch()

        kitsu_button = QtWidgets.QPushButton(QtGui.QIcon(resources.get_icon(('icons.gui', 'kitsu'))), '')
        kitsu_button.setObjectName('ProjectButton')
        kitsu_button.clicked.connect(self._on_kitsu_button_clicked)

        self.start_button = QtWidgets.QPushButton(
            QtGui.QIcon(resources.get_icon(('icons.gui', 'start'), color=QtGui.QColor('white'))), ''
        )
        self.start_button.setObjectName('ProjectButton')
        self.start_button.clicked.connect(self._on_start_button_clicked)

        self.start_button.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.start_button.customContextMenuRequested.connect(self._on_context_menu)

        footer_lo.addWidget(kitsu_button, alignment=QtCore.Qt.AlignVCenter)
        footer_lo.addWidget(self.start_button, alignment=QtCore.Qt.AlignVCenter)

    def refresh(self, data=None):
        if data:
            self._status = data['updated']
            self._data = data['data']

        if self._status is True:
            color = QtGui.QColor('#4eb16f')
            self.status_lbl.setToolTip("Ready to launch")
        elif self._status is False:
            color = QtGui.QColor('#d5972a')
            self.status_lbl.setToolTip("Needs to be updated")
        else:
            color = QtGui.QColor('#d5972a')
            self.status_lbl.setToolTip("Waiting for first install")

        self.circular_shape = QtGui.QIcon(
            resources.get_icon(('icons.gui', 'circular-shape-silhouette'), color=color)
        )
        pixmap = self.circular_shape.pixmap(QtCore.QSize(24, 24))
        self.status_lbl.setPixmap(pixmap)

        self.project_name.setText(self._data['name'])

    def _on_kitsu_button_clicked(self):
        webbrowser.open(self._data['kitsu_url'])

    def _on_start_button_clicked(self):
        data_resolve = None
        if not self._status:
            updated_data = self._ctrl.update_project(self._uid)

            print(f"INFO: Updating {self._data['name']} libreflow")
            self._mw.popup.pop('updating')
            QtWidgets.QApplication.processEvents()
            
            data_resolve, env_folder = self._ctrl.update_project_env(self._data['name'], self._uid)
            if data_resolve is None and env_folder is None:
                self._mw.notification_header.pop("Error has occurred. Check console.", 'error', True)
                self._mw.popup.hide()
                return
            
            updated_data = self._ctrl.update_project(self._uid, updated=True)
            
            self.refresh(updated_data)

        # self._mw.popup.pop('launching')
        # QtWidgets.QApplication.processEvents()

        print(f"INFO: Start {self._data['name']} libreflow")
        self._ctrl.start_project(self._uid, data_resolve)

        self._mw.popup.hide()

    def _on_context_menu(self, event):
        context_menu = QtWidgets.QMenu(self)

        update_action = QtWidgets.QAction('Force update', self)
        update_action.triggered.connect(self._on_force_update_clicked)

        context_menu.addAction(update_action)

        context_menu.exec_(self.start_button.mapToGlobal(event))

    def _on_force_update_clicked(self):
        print(f"INFO: Updating {self._data['name']} libreflow")
        self._mw.popup.pop('updating')
        QtWidgets.QApplication.processEvents()
        
        data_resolve, env_folder = self._ctrl.update_project_env(self._data['name'], self._uid)
        if data_resolve is None and env_folder is None:
            self._mw.notification_header.pop("Error has occurred. Check console.", 'error', True)
            self._mw.popup.hide()
            return

        updated_data = self._ctrl.update_project(self._uid, updated=True)    
        self.refresh(updated_data)

        print(f"INFO: Start {self._data['name']} libreflow")
        self._ctrl.start_project(self._uid, data_resolve)

        self._mw.popup.hide()


class ProjectsList(QtWidgets.QTableWidget):

    def __init__(self, mw):
        super(ProjectsList, self).__init__()
        self._mw = mw
        self._ctrl = mw._ctrl

        self.cellsContent = []

        ### For grid view
        # self.cellWidth = 335
        # self.cellHeight = 200
        self.cellPadding = 15
        
        ### For grid view
        # self.currentColumnCount = self.viewport().width() // self.cellWidth
        # self.setColumnCount(self.currentColumnCount)
        self.setColumnCount(1)

        self.setShowGrid(False)
        self.horizontalHeader().setVisible(False)
        self.verticalHeader().setVisible(False)
        self.setVerticalScrollMode(QtWidgets.QAbstractItemView.ScrollPerPixel)

        self.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self.setFocusPolicy(QtCore.Qt.NoFocus)
        self.setSelectionMode(QtWidgets.QAbstractItemView.NoSelection)

        self.setStyleSheet(
            f'''
            QTableWidget {{
                background-color: transparent;
                border: none;
            }}
            '''
        )
    
    def refresh(self):
        self.blockSignals(True)

        if self.count() > 0:
            for cell in self.cellsContent:
                self.removeCellWidget(cell[0], cell[1])
            
            self.cellsContent = []

        projects = self._ctrl.fetch_projects()
        projects = sorted(projects, key=lambda p: p['data']['name']) 

        for project in projects:
            start_r, start_c = self.get_last_idx()
            if start_c == 0:
                self.setRowCount(start_r + 1)

            if project['server_uid'] == self._ctrl.server_uid:
                if project['data']['is_enabled'] and not project['data']['is_archived']:
                    item = ProjectItem(self, project)
                    lo = QtWidgets.QHBoxLayout()
                    lo.addWidget(item)
                    lo.setContentsMargins(self.cellPadding, self.cellPadding, self.cellPadding, self.cellPadding)

                    widget = QtWidgets.QWidget()
                    widget.setLayout(lo)
                    self.setCellWidget(start_r, start_c, widget)
                    self.setRowHeight(start_r, widget.height())

                    self.cellsContent.append((start_r, start_c, widget))
        
        self.blockSignals(False)

    def get_last_idx(self):
        last_r, last_c = divmod(self.count(), self.columnCount())
        return last_r, last_c

    def count(self):
        return len(self.cellsContent)

    def resizeEvent(self, event):
        ### For grid view
        ### TODO: Find why it crashes after a project was added or deleted
        # possibleColumn = self.viewport().width() // 335
        # if possibleColumn == 0:
        #     possibleColumn = 1
        
        # # Reorder widgets
        # if possibleColumn != self.currentColumnCount:
        #     # Remove all widgets
        #     for cell in self.cellsContent:
        #         self.removeCellWidget(cell[0], cell[1])

        #     # Set new column count
        #     self.setColumnCount(possibleColumn)
        #     self.currentColumnCount = possibleColumn

        #     # Re-add widgets
        #     for c, cell in enumerate(self.cellsContent):
        #         row_index, column_index, widget = cell[0], cell[1], cell[2]

        #         start_r, start_c = divmod(c, self.columnCount())
        #         if start_c == 0:
        #             self.setRowCount(start_r + 1)

        #         self.setCellWidget(start_r, start_c, widget)
        #         self.cellsContent[c] = (start_r, start_c, widget)
       
        # Set column width and row height
        ### Fixed row height for grid view
        ### Fixed row height based on widget height itself
        # column_width = self.viewport().width() // self.currentColumnCount
        column_width = self.viewport().width() // 1
        # for r_idx in range(self.rowCount()):
        #     self.setRowHeight(r_idx, self.cellHeight)
        for c_idx in range(self.columnCount()):
            self.setColumnWidth(c_idx, column_width)

        return super().resizeEvent(event)
